# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import afsmon
import configparser

from afsmon.cmd.main import AFSMonCmd
from afsmon.tests import base

"""
test_afsmon
----------------------------------

Tests for `afsmon` module.
"""


class TestPyAFSMon(base.TestCase):

    def setUp(self):
        super(TestPyAFSMon, self).setUp()

    def test_statsd(self):
        cmd = AFSMonCmd()
        cmd.config = configparser.ConfigParser()

        a = afsmon.FileServerStats('afs01.dfw.openstack.org')
        a.status = afsmon.FileServerStatus.NORMAL
        a.idle_threads = 250
        a.calls_waiting = 0
        a.partitions = [afsmon.Partition('vicepa', 512, 512, 1024, 50.00)]
        a.volumes = [
            afsmon.Volume('mirror.foo', 12345678, 'RW', 512, 1024, 50.00),
            afsmon.Volume('mirror.moo', 87654321, 'RW', 1024, 2048, 50.00),
        ]

        b = afsmon.FileServerStats('afs02.ord.openstack.org')
        b.status = afsmon.FileServerStatus.NORMAL
        b.idle_threads = 100
        b.calls_waiting = 2
        b.partitions = [afsmon.Partition('vicepa', 512, 512, 1024, 50.00)]
        b.volumes = []

        cmd.fileservers = [a, b]

        cmd.cmd_statsd()

        self.assertReportedStat(
            'afs.afs01_dfw_openstack_org.idle_threads', value='250', kind='g')
        self.assertReportedStat(
            'afs.afs02_ord_openstack_org.calls_waiting', value='2', kind='g')
        self.assertReportedStat(
            'afs.afs01_dfw_openstack_org.part.vicepa.used',
            value='512', kind='g')
        self.assertReportedStat(
            'afs.afs01_dfw_openstack_org.part.vicepa.total',
            value='1024', kind='g')
        self.assertReportedStat(
            'afs.afs01_dfw_openstack_org.vol.mirror_moo.used',
            value='1024', kind='g')
        self.assertReportedStat(
            'afs.afs01_dfw_openstack_org.vol.mirror_moo.quota',
            value='2048', kind='g')



#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import argparse
import configparser
import logging
import os
import statsd
import sys

import afsmon

logger = logging.getLogger("afsmon.main")


class AFSMonCmd(object):

    def cmd_show(self):
        for fs in self.fileservers:
            print(fs)
        return 0

    def cmd_statsd(self):
        # note we're just being careful to let the default values fall
        # through to StatsClient()
        statsd_args = {}
        try:
            try:
                statsd_args['host'] = self.config.get('statsd', 'host')
            except configparser.NoOptionError:
                pass
            try:
                statsd_args['port'] = self.config.get('statsd', 'port')
            except configparser.NoOptionerror:
                pass
        except configparser.NoSectionError:
            pass
        if os.getenv('STATSD_HOST', None):
            statsd_args['host'] = os.environ['STATSD_HOST']
        if os.getenv('STATSD_PORT', None):
            statsd_args['port'] = os.environ['STATSD_PORT']
        logger.debug("Sending stats to %s:%s" % (statsd_args['host'],
                                                 statsd_args['port']))
        self.statsd = statsd.StatsClient(**statsd_args)

        for f in self.fileservers:
            if f.status != afsmon.FileServerStatus.NORMAL:
                continue

            hn = f.hostname.replace('.', '_')
            self.statsd.gauge('afs.%s.idle_threads' % hn, f.idle_threads)
            self.statsd.gauge('afs.%s.calls_waiting' % hn, f.calls_waiting)
            for p in f.partitions:
                self.statsd.gauge(
                    'afs.%s.part.%s.used' % (hn, p.partition), p.used)
                self.statsd.gauge(
                    'afs.%s.part.%s.free' % (hn, p.partition), p.free)
                self.statsd.gauge(
                    'afs.%s.part.%s.total' % (hn, p.partition), p.total)
            for v in f.volumes:
                if v.perms != 'RW':
                    continue
                vn = v.volume.replace('.', '_')
                self.statsd.gauge(
                    'afs.%s.vol.%s.used' % (hn, vn), v.used)
                self.statsd.gauge(
                    'afs.%s.vol.%s.quota' % (hn, vn), v.quota)

    def main(self, args=None):
        if args is None:
            args = sys.argv[1:]

        self.fileservers = []

        parser = argparse.ArgumentParser(
            description='An AFS monitoring tool')

        parser.add_argument("-c", "--config", action='store',
                            default="/etc/afsmon.cfg",
                            help="Path to config file")
        parser.add_argument("-d", '--debug', action="store_true")

        subparsers = parser.add_subparsers(title='commands',
                                           description='valid commands',
                                           dest='command')

        cmd_show = subparsers.add_parser('show', help='show table of results')
        cmd_show.set_defaults(func=self.cmd_show)

        cmd_statsd = subparsers.add_parser('statsd', help='report to statsd')
        cmd_statsd.set_defaults(func=self.cmd_statsd)

        self.args = parser.parse_args(args)

        if self.args.debug:
            logging.basicConfig(level=logging.DEBUG)
            logger.debug("Debugging enabled")

        if not os.path.exists(self.args.config):
            raise ValueError("Config file %s does not exist" %
                             self.args.config)

        self.config = configparser.RawConfigParser()
        self.config.read(self.args.config)

        cell = self.config.get('main', 'cell').strip()

        fs_addrs = afsmon.get_fs_addresses(cell)
        logger.debug("Found fileservers: %s" % ", ".join(fs_addrs))

        for addr in fs_addrs:
            logger.debug("Finding stats for: %s" % addr)
            fs = afsmon.FileServerStats(addr)
            fs.get_stats()
            self.fileservers.append(fs)

        # run the subcommand
        return self.args.func()


def main():
    cmd = AFSMonCmd()
    return cmd.main()



#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import collections
import io
import logging
import re
import subprocess


from datetime import datetime
from enum import Enum
from prettytable import PrettyTable


logger = logging.getLogger("afsmon")


class FileServerStatus(Enum):
    NORMAL = 0
    TEMPORARY_DISABLED = 1
    DISABLED = 2
    UNKNOWN = 3
    NO_CONNECTION = 4

Partition = collections.namedtuple(
    'Partition', 'partition, used, free, total, percent_used')

Volume = collections.namedtuple(
    'Voume', 'volume, id, perms, used, quota, percent_used')


class FileServerStats(object):
    '''AFS fileserver status

    Call ``get_stats()`` to populate the statistics for the server.
    Note most attributes are only set if ``status`` is NORMAL

    Args:
       hostname (str): The hostname of server to query
         i.e. argument to ``-server`` for cmd line tools

    Attributes:
       status (FileServerStatus): enum of possible status
       timestamp(:obj:`datetime.datetime`): time statistics retrieved
       restart (:obj:`datetime.datetime`): time of last restart
       uptime (:obj:`datetime.timedelta`): current uptime
       partitions (:obj:`list`): list of :obj:`Partition` tuples for each
         partition on the server
       calls_waiting (:obj:`int`): number of calls waiting for a thread
       idle_threads (:obj:`int`): number of currently idle threads
       volumes (:obj:`list`): list of :obj:`Volume` tuples for each
         volume present on the server
       table (:obj:`PrettyTable`): a printable PrettyTable representation

    '''

    def _get_volumes(self):
        cmd = ["vos", "listvol", "-long", "-server", self.hostname]
        logger.debug("Running: %s" % cmd)
        output = subprocess.check_output(
            cmd, stderr=subprocess.STDOUT).decode('ascii')

        # Matching:
        # mirror.yum-puppetlabs.readonly    536871036 RO   63026403 K  On-line
        vol_regex = re.compile(
          '^(?P<vol>[^\s]+)\s+(?P<id>\d+)\s(?P<perms>R[OW])\s+(?P<used>\d+) K'
        )

        # Read the output into chunks where each chunk is the info for
        # one volume.
        lines = io.StringIO(output)
        while True:
            line = lines.readline()
            if not line:
                break
            chunk = ''
            if "On-line" in line:  # chunks start with this
                chunk += line
                # read in the next 9 lines of status
                for i in range(8):
                    chunk += lines.readline()
                # convert it to a Volume()
                # todo: there's a bunch more we could extract...
                m = vol_regex.search(chunk)
                q = re.search('MaxQuota\s+(?P<quota>\d+) K', chunk)
                used = int(m['used'])
                quota = int(q['quota'])
                percent_used = round(float(used) / float(quota) * 100, 2)
                self.volumes.append(
                    Volume(m['vol'], m['id'], m['perms'],
                           used, quota, percent_used))

    def _get_calls_waiting(self):
        cmd = ["rxdebug", self.hostname, "7000", "-rxstats", "-noconns"]
        logger.debug("Running: %s" % cmd)
        output = subprocess.check_output(
            cmd, stderr=subprocess.STDOUT).decode('ascii')

        for line in output.split('\n'):
            m = re.search('(?P<waiting>\d+) calls waiting for a thread', line)
            if m:
                self.calls_waiting = int(m['waiting'])
            m = re.search('(?P<idle>\d+) threads are idle', line)
            if m:
                self.idle_threads = int(m['idle'])

    def _get_partition_stats(self):
        cmd = ["vos", "partinfo", self.hostname, "-noauth"]
        logger.debug("Running: %s" % cmd)
        output = subprocess.check_output(
            cmd, stderr=subprocess.STDOUT).decode('ascii')

        for line in output.split('\n'):
            m = re.search(
                'Free space on partition '
                '/vicep(?P<partition>[a-z][a-z]?): '
                '(?P<free>\d+) K blocks out of total (?P<total>\d+)', line)
            if m:
                part = 'vicep%s' % m['partition']
                # (used, free, total, %age)
                used = int(m['total']) - int(m['free'])
                self.partitions.append(
                    Partition(part, used, int(m['free']), int(m['total']),
                              round(float(used) / float(m['total']) * 100, 2)))

    def _get_fs_stats(self):
        cmd = ["bos", "status", self.hostname, "-long", "-noauth"]
        logger.debug("Running: %s" % cmd)
        try:
            output = subprocess.check_output(
                cmd, stderr=subprocess.STDOUT).decode('ascii')
        except subprocess.CalledProcessError:
            logger.debug(" ... failed!")
            self.status = FileServerStatus.NO_CONNECTION
            return

        if re.search('currently running normally', output):
            self.status = FileServerStatus.NORMAL
            m = re.search(
                r'last started at (?P<date>\w+ \w+ \w+ \d+:\d+:\d+ \d+)',
                output)
            self.restart = datetime.strptime(m['date'], '%a %b %d %H:%M:%S %Y')
            self.uptime = self.timestamp - self.restart

        elif re.search('temporarily disabled, currently shutdown', output):
            self.status = FileServerStatus.TEMPORARILY_DISABLED
        elif re.search('disabled, currently shutdown', output):
            self.status = FileServerStatus.DISABLED
        else:
            logger.debug(output)
            self.status = FileServerStatus.UNKNOWN

    def get_stats(self):
        '''Get the complete stats set for the fileserver'''
        self.timestamp = datetime.now()

        self._get_fs_stats()
        if self.status == FileServerStatus.NORMAL:
            self._get_partition_stats()
            self._get_calls_waiting()
            self._get_volumes()

        self.table = PrettyTable()
        self.table.field_names = ["Metric", "Value"]
        self.table.align["Metric"] = "l"
        self.table.align["Value"] = "l"
        self.table.add_row(["Hostname", self.hostname])
        self.table.add_row(["Timestamp", self.timestamp])
        self.table.add_row(["Status", self.status])
        self.table.add_row(["Uptime", self.uptime])
        self.table.add_row(["Last Restart", self.restart])
        self.table.add_row(["Calls Waiting", self.calls_waiting])
        self.table.add_row(["Idle Threads", self.idle_threads])
        for p in self.partitions:
            n = "/%s" % p.partition
            self.table.add_row(["%s used" % n, p.used])
            self.table.add_row(["%s free" % n, p.free])
            self.table.add_row(["%s total" % n, p.total])
            self.table.add_row(["%s %%used" % n,
                                "%s%%" % p.percent_used])
        for v in self.volumes:
            # Only add the RW volumes to the table as for now we're
            # mostly just worried about viewing the quota.
            if v.perms == 'RW':
                n = v.volume
                self.table.add_row(["%s used" % n, v.used])
                self.table.add_row(["%s quota" % n, v.quota])
                self.table.add_row(["%s %%used" % n,
                                    "%s%%" % v.percent_used])

    def __str__(self):
        return str(self.table)

    def __init__(self, hostname):
        self.hostname = hostname

        self.timestamp = None
        self.restart = None
        self.uptime = None
        self.partitions = []
        self.volumes = []
        self.calls_waiting = None
        self.idle_threads = None


def get_fs_addresses(cell):
    '''Get the fileservers associated with a cell

    :arg str cell: The cell (e.g. ``openstack.org``)
    :returns: list of fileservers for the cell
    '''
    fs = []
    cmd = ["vos", "listaddrs", "-noauth", "-cell", cell]
    logger.debug("Running: %s" % cmd)
    try:
        output = subprocess.check_output(
            cmd, stderr=subprocess.STDOUT).decode('ascii')
    except subprocess.CalledProcessError:
        logger.debug(" ... failed!")
        return []

    for line in output.split('\n'):
        if line.strip():
            fs.append(line)

    return fs



# Copyright 2010-2011 OpenStack Foundation
# Copyright (c) 2013 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import fixtures
import logging
import os
import select
import socket
import testtools
import threading
import time


_TRUE_VALUES = ('True', 'true', '1', 'yes')

logger = logging.getLogger("afsmon.tests.base")


class FakeStatsd(threading.Thread):
    def __init__(self):
        threading.Thread.__init__(self)
        self.daemon = True
        self.sock = socket.socket(socket.AF_INET6, socket.SOCK_DGRAM)
        self.sock.bind(('', 0))
        self.port = self.sock.getsockname()[1]
        self.wake_read, self.wake_write = os.pipe()
        self.stats = []

    def run(self):
        while True:
            poll = select.poll()
            poll.register(self.sock, select.POLLIN)
            poll.register(self.wake_read, select.POLLIN)
            ret = poll.poll()
            for (fd, event) in ret:
                if fd == self.sock.fileno():
                    data = self.sock.recvfrom(1024)
                    if not data:
                        return
                    self.stats.append(data[0])
                if fd == self.wake_read:
                    return

    def stop(self):
        os.write(self.wake_write, b'1\n')


class TestCase(testtools.TestCase):
    """Test case base class for all unit tests."""

    def setUp(self):
        """Run before each test method to initialize test environment."""

        super(TestCase, self).setUp()
        test_timeout = os.environ.get('OS_TEST_TIMEOUT', 0)
        try:
            test_timeout = int(test_timeout)
        except ValueError:
            # If timeout value is invalid do not set a timeout.
            test_timeout = 0
        if test_timeout > 0:
            self.useFixture(fixtures.Timeout(test_timeout, gentle=True))

        self.useFixture(fixtures.NestedTempfile())
        self.useFixture(fixtures.TempHomeDir())

        if os.environ.get('OS_STDOUT_CAPTURE') in _TRUE_VALUES:
            stdout = self.useFixture(fixtures.StringStream('stdout')).stream
            self.useFixture(fixtures.MonkeyPatch('sys.stdout', stdout))
        if os.environ.get('OS_STDERR_CAPTURE') in _TRUE_VALUES:
            stderr = self.useFixture(fixtures.StringStream('stderr')).stream
            self.useFixture(fixtures.MonkeyPatch('sys.stderr', stderr))

        self.log_fixture = self.useFixture(
            fixtures.FakeLogger(level=logging.DEBUG))

        self.statsd = FakeStatsd()
        os.environ['STATSD_HOST'] = '127.0.0.1'
        os.environ['STATSD_PORT'] = str(self.statsd.port)
        self.statsd.start()

    def shutdown(self):
        self.statsd.stop()
        self.statsd.join()

    def assertReportedStat(self, key, value=None, kind=None):
        """Check statsd output

        Check statsd return values.  A ``value`` should specify a
        ``kind``, however a ``kind`` may be specified without a
        ``value`` for a generic match.  Leave both empy to just check
        for key presence.

        :arg str key: The statsd key
        :arg str value: The expected value of the metric ``key``
        :arg str kind: The expected type of the metric ``key``  For example

          - ``c`` counter
          - ``g`` gauge
          - ``ms`` timing
          - ``s`` set
        """
        if value:
            self.assertNotEqual(kind, None)

        start = time.time()
        while time.time() < (start + 5):
            # Note our fake statsd just queues up results in a queue.
            # We just keep going through them until we find one that
            # matches, or fail out.
            for stat in self.statsd.stats:
                k, v = stat.decode('utf-8').split(':')
                if key == k:
                    if kind is None:
                        # key with no qualifiers is found
                        return True

                    s_value, s_kind = v.split('|')
                    # if no kind match, look for other keys
                    if kind != s_kind:
                        continue

                    if value:
                        # special-case value|ms because statsd can turn
                        # timing results into float of indeterminate
                        # length, hence foiling string matching.
                        if kind == 'ms':
                            if float(value) == float(s_value):
                                return True
                        if value == s_value:
                            return True
                        # otherwise keep looking for other matches
                        continue

                    # this key matches
                    return True
            time.sleep(0.1)

        raise Exception("Key %s not found in reported stats" % key)
